@@ -7,7 +7,7 @@ goes into the "Main" module for the individual binaries.
 
 {-
 
-Copyright (C) 2009 Google Inc.
+Copyright (C) 2009, 2010, 2011 Google Inc.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 module Ganeti.HTools.Cluster
     (
      -- * Types
-      AllocSolution
+      AllocSolution(..)
     , Table(..)
     , CStats(..)
     , AllocStats
@@ -51,26 +51,39 @@ module Ganeti.HTools.Cluster
     , doNextBalance
     , tryBalance
     , compCV
+    , compDetailedCV
     , printStats
     , iMoveToJob
     -- * IAllocator functions
+    , genAllocNodes
     , tryAlloc
+    , tryMGAlloc
     , tryReloc
+    , tryMGReloc
     , tryEvac
+    , tryMGEvac
     , collapseFailures
     -- * Allocation functions
     , iterateAlloc
     , tieredAlloc
+    , tieredSpecMap
+     -- * Node group functions
+    , instanceGroup
+    , findSplitInstances
+    , splitCluster
     ) where
 
+import Data.Function (on)
 import Data.List
 import Data.Ord (comparing)
 import Text.Printf (printf)
 import Control.Monad
+import Control.Parallel.Strategies
 
 import qualified Ganeti.HTools.Container as Container
 import qualified Ganeti.HTools.Instance as Instance
 import qualified Ganeti.HTools.Node as Node
+import qualified Ganeti.HTools.Group as Group
 import Ganeti.HTools.Types
 import Ganeti.HTools.Utils
 import qualified Ganeti.OpCodes as OpCodes
@@ -78,11 +91,34 @@ import qualified Ganeti.OpCodes as OpCodes
 -- * Types
 
 -- | Allocation\/relocation solution.
-type AllocSolution = ([FailMode], Int, [(Score, Node.AllocElement)])
+data AllocSolution = AllocSolution
+  { asFailures  :: [FailMode]          -- ^ Failure counts
+  , asAllocs    :: Int                 -- ^ Good allocation count
+  , asSolutions :: [Node.AllocElement] -- ^ The actual result, length
+                                       -- of the list depends on the
+                                       -- allocation/relocation mode
+  , asLog       :: [String]            -- ^ A list of informational messages
+  }
+
+-- | Allocation results, as used in 'iterateAlloc' and 'tieredAlloc'.
+type AllocResult = (FailStats, Node.List, Instance.List,
+                    [Instance.Instance], [CStats])
+
+
+-- | A type denoting the valid allocation mode/pairs.
+-- For a one-node allocation, this will be a @Left ['Node.Node']@,
+-- whereas for a two-node allocation, this will be a @Right
+-- [('Node.Node', 'Node.Node')]@.
+type AllocNodes = Either [Ndx] [(Ndx, Ndx)]
+
+-- | The empty solution we start with when computing allocations
+emptySolution :: AllocSolution
+emptySolution = AllocSolution { asFailures = [], asAllocs = 0
+                              , asSolutions = [], asLog = [] }
 
 -- | The complete state for the balancing solution
 data Table = Table Node.List Instance.List Score [Placement]
-             deriving (Show)
+             deriving (Show, Read)
 
 data CStats = CStats { csFmem :: Int    -- ^ Cluster free mem
                      , csFdsk :: Int    -- ^ Cluster free disk
@@ -106,7 +142,7 @@ data CStats = CStats { csFmem :: Int    -- ^ Cluster free mem
                      , csScore :: Score -- ^ The cluster score
                      , csNinst :: Int   -- ^ The total number of instances
                      }
-            deriving (Show)
+            deriving (Show, Read)
 
 -- | Currently used, possibly to allocate, unallocable
 type AllocStats = (RSpec, RSpec, RSpec)
@@ -158,15 +194,16 @@ updateCStats cs node =
         inc_icpu = Node.uCpu node
         inc_idsk = truncate (Node.tDsk node) - Node.fDsk node
         inc_vcpu = Node.hiCpu node
+        inc_acpu = Node.availCpu node
 
     in cs { csFmem = x_fmem + Node.fMem node
           , csFdsk = x_fdsk + Node.fDsk node
           , csAmem = x_amem + inc_amem'
           , csAdsk = x_adsk + inc_adsk
-          , csAcpu = x_acpu
+          , csAcpu = x_acpu + inc_acpu
           , csMmem = max x_mmem inc_amem'
           , csMdsk = max x_mdsk inc_adsk
-          , csMcpu = x_mcpu
+          , csMcpu = max x_mcpu inc_acpu
           , csImem = x_imem + inc_imem
           , csIdsk = x_idsk + inc_idsk
           , csIcpu = x_icpu + inc_icpu
@@ -215,7 +252,7 @@ detailedCVInfo = [ (1,  "free_mem_cv")
                  , (1,  "mem_load_cv")
                  , (1,  "disk_load_cv")
                  , (1,  "net_load_cv")
-                 , (1,  "pri_tags_score")
+                 , (2,  "pri_tags_score")
                  ]
 
 detailedCVWeights :: [Double]
@@ -230,16 +267,16 @@ compDetailedCV nl =
         mem_l = map Node.pMem nodes
         dsk_l = map Node.pDsk nodes
         -- metric: memory covariance
-        mem_cv = varianceCoeff mem_l
+        mem_cv = stdDev mem_l
         -- metric: disk covariance
-        dsk_cv = varianceCoeff dsk_l
+        dsk_cv = stdDev dsk_l
         -- metric: count of instances living on N1 failing nodes
         n1_score = fromIntegral . sum . map (\n -> length (Node.sList n) +
                                                    length (Node.pList n)) .
                    filter Node.failN1 $ nodes :: Double
         res_l = map Node.pRem nodes
         -- metric: reserved memory covariance
-        res_cv = varianceCoeff res_l
+        res_cv = stdDev res_l
         -- offline instances metrics
         offline_ipri = sum . map (length . Node.pList) $ offline
         offline_isec = sum . map (length . Node.sList) $ offline
@@ -251,7 +288,7 @@ compDetailedCV nl =
         off_pri_score = fromIntegral offline_ipri::Double
         cpu_l = map Node.pCpu nodes
         -- metric: covariance of vcpu/pcpu ratio
-        cpu_cv = varianceCoeff cpu_l
+        cpu_cv = stdDev cpu_l
         -- metrics: covariance of cpu, memory, disk and network load
         (c_load, m_load, d_load, n_load) = unzip4 $
             map (\n ->
@@ -263,8 +300,7 @@ compDetailedCV nl =
         pri_tags_inst = sum $ map Node.conflictingPrimaries nodes
         pri_tags_score = fromIntegral pri_tags_inst::Double
     in [ mem_cv, dsk_cv, n1_score, res_cv, off_score, off_pri_score, cpu_cv
-       , varianceCoeff c_load, varianceCoeff m_load
-       , varianceCoeff d_load, varianceCoeff n_load
+       , stdDev c_load, stdDev m_load , stdDev d_load, stdDev n_load
        , pri_tags_score ]
 
 -- | Compute the /total/ variance.
@@ -379,28 +415,28 @@ applyMove nl inst (FailoverAndReplace new_sdx) =
     in new_nl
 
 -- | Tries to allocate an instance on one given node.
-allocateOnSingle :: Node.List -> Instance.Instance -> Node.Node
+allocateOnSingle :: Node.List -> Instance.Instance -> Ndx
                  -> OpResult Node.AllocElement
-allocateOnSingle nl inst p =
-    let new_pdx = Node.idx p
+allocateOnSingle nl inst new_pdx =
+    let p = Container.find new_pdx nl
         new_inst = Instance.setBoth inst new_pdx Node.noSecondary
-        new_nl = Node.addPri p inst >>= \new_p ->
-                 return (Container.add new_pdx new_p nl, new_inst, [new_p])
-    in new_nl
+    in  Node.addPri p inst >>= \new_p -> do
+      let new_nl = Container.add new_pdx new_p nl
+          new_score = compCV nl
+      return (new_nl, new_inst, [new_p], new_score)
 
 -- | Tries to allocate an instance on a given pair of nodes.
-allocateOnPair :: Node.List -> Instance.Instance -> Node.Node -> Node.Node
+allocateOnPair :: Node.List -> Instance.Instance -> Ndx -> Ndx
                -> OpResult Node.AllocElement
-allocateOnPair nl inst tgt_p tgt_s =
-    let new_pdx = Node.idx tgt_p
-        new_sdx = Node.idx tgt_s
-        new_nl = do -- Maybe monad
-          new_p <- Node.addPri tgt_p inst
-          new_s <- Node.addSec tgt_s inst new_pdx
-          let new_inst = Instance.setBoth inst new_pdx new_sdx
-          return (Container.addTwo new_pdx new_p new_sdx new_s nl, new_inst,
-                 [new_p, new_s])
-    in new_nl
+allocateOnPair nl inst new_pdx new_sdx =
+    let tgt_p = Container.find new_pdx nl
+        tgt_s = Container.find new_sdx nl
+    in do
+      new_p <- Node.addPri tgt_p inst
+      new_s <- Node.addSec tgt_s inst new_pdx
+      let new_inst = Instance.setBoth inst new_pdx new_sdx
+          new_nl = Container.addTwo new_pdx new_p new_sdx new_s nl
+      return (new_nl, new_inst, [new_p, new_s], compCV new_nl)
 
 -- | Tries to perform an instance move and returns the best table
 -- between the original one and the new one.
@@ -471,13 +507,18 @@ checkMove :: [Ndx]               -- ^ Allowed target node indices
           -> Table               -- ^ The new solution
 checkMove nodes_idx disk_moves ini_tbl victims =
     let Table _ _ _ ini_plc = ini_tbl
+        -- we're using rwhnf from the Control.Parallel.Strategies
+        -- package; we don't need to use rnf as that would force too
+        -- much evaluation in single-threaded cases, and in
+        -- multi-threaded case the weak head normal form is enough to
+        -- spark the evaluation
+        tables = parMap rwhnf (checkInstanceMove nodes_idx disk_moves ini_tbl)
+                 victims
         -- iterate over all instances, computing the best move
         best_tbl =
             foldl'
-            (\ step_tbl em ->
-                 compareTables step_tbl $
-                 checkInstanceMove nodes_idx disk_moves ini_tbl em)
-            ini_tbl victims
+            (\ step_tbl new_tbl -> compareTables step_tbl new_tbl)
+            ini_tbl tables
         Table _ _ _ best_plc = best_tbl
     in if length best_plc == length ini_plc
        then ini_tbl -- no advancement
@@ -497,8 +538,10 @@ doNextBalance ini_tbl max_rounds min_score =
 tryBalance :: Table       -- ^ The starting table
            -> Bool        -- ^ Allow disk moves
            -> Bool        -- ^ Only evacuate moves
+           -> Score       -- ^ Min gain threshold
+           -> Score       -- ^ Min gain
            -> Maybe Table -- ^ The resulting table and commands
-tryBalance ini_tbl disk_moves evac_mode =
+tryBalance ini_tbl disk_moves evac_mode mg_limit min_gain =
     let Table ini_nl ini_il ini_cv _ = ini_tbl
         all_inst = Container.elems ini_il
         all_inst' = if evac_mode
@@ -514,7 +557,7 @@ tryBalance ini_tbl disk_moves evac_mode =
         fin_tbl = checkMove node_idx disk_moves ini_tbl reloc_inst
         (Table _ _ fin_cv _) = fin_tbl
     in
-      if fin_cv < ini_cv
+      if fin_cv < ini_cv && (ini_cv > mg_limit || ini_cv - fin_cv >= min_gain)
       then Just fin_tbl -- this round made success, return the new table
       else Nothing
 
@@ -528,57 +571,158 @@ collapseFailures flst =
 -- | Update current Allocation solution and failure stats with new
 -- elements
 concatAllocs :: AllocSolution -> OpResult Node.AllocElement -> AllocSolution
-concatAllocs (flst, cntok, sols) (OpFail reason) = (reason:flst, cntok, sols)
+concatAllocs as (OpFail reason) = as { asFailures = reason : asFailures as }
 
-concatAllocs (flst, cntok, osols) (OpGood ns@(nl, _, _)) =
-    let nscore = compCV nl
-        -- Choose the old or new solution, based on the cluster score
+concatAllocs as (OpGood ns@(_, _, _, nscore)) =
+    let -- Choose the old or new solution, based on the cluster score
+        cntok = asAllocs as
+        osols = asSolutions as
         nsols = case osols of
-                  [] -> [(nscore, ns)]
-                  (oscore, _):[] ->
+                  [] -> [ns]
+                  (_, _, _, oscore):[] ->
                       if oscore < nscore
                       then osols
-                      else [(nscore, ns)]
+                      else [ns]
                   -- FIXME: here we simply concat to lists with more
                   -- than one element; we should instead abort, since
                   -- this is not a valid usage of this function
-                  xs -> (nscore, ns):xs
+                  xs -> ns:xs
         nsuc = cntok + 1
     -- Note: we force evaluation of nsols here in order to keep the
     -- memory profile low - we know that we will need nsols for sure
     -- in the next cycle, so we force evaluation of nsols, since the
     -- foldl' in the caller will only evaluate the tuple, but not the
     -- elements of the tuple
-    in nsols `seq` nsuc `seq` (flst, nsuc, nsols)
+    in nsols `seq` nsuc `seq` as { asAllocs = nsuc, asSolutions = nsols }
+
+-- | Sums two allocation solutions (e.g. for two separate node groups).
+sumAllocs :: AllocSolution -> AllocSolution -> AllocSolution
+sumAllocs (AllocSolution af aa as al) (AllocSolution bf ba bs bl) =
+    AllocSolution (af ++ bf) (aa + ba) (as ++ bs) (al ++ bl)
+
+-- | Given a solution, generates a reasonable description for it
+describeSolution :: AllocSolution -> String
+describeSolution as =
+  let fcnt = asFailures as
+      sols = asSolutions as
+      freasons =
+        intercalate ", " . map (\(a, b) -> printf "%s: %d" (show a) b) .
+        filter ((> 0) . snd) . collapseFailures $ fcnt
+  in if null sols
+     then "No valid allocation solutions, failure reasons: " ++
+          (if null fcnt
+           then "unknown reasons"
+           else freasons)
+     else let (_, _, nodes, cv) = head sols
+          in printf ("score: %.8f, successes %d, failures %d (%s)" ++
+                     " for node(s) %s") cv (asAllocs as) (length fcnt) freasons
+             (intercalate "/" . map Node.name $ nodes)
+
+-- | Annotates a solution with the appropriate string
+annotateSolution :: AllocSolution -> AllocSolution
+annotateSolution as = as { asLog = describeSolution as : asLog as }
+
+-- | Generate the valid node allocation singles or pairs for a new instance.
+genAllocNodes :: Group.List        -- ^ Group list
+              -> Node.List         -- ^ The node map
+              -> Int               -- ^ The number of nodes required
+              -> Bool              -- ^ Whether to drop or not
+                                   -- unallocable nodes
+              -> Result AllocNodes -- ^ The (monadic) result
+genAllocNodes gl nl count drop_unalloc =
+    let filter_fn = if drop_unalloc
+                    then filter ((/=) AllocUnallocable . Group.allocPolicy .
+                                     flip Container.find gl . Node.group)
+                    else id
+        all_nodes = filter_fn $ getOnline nl
+        all_pairs = liftM2 (,) all_nodes all_nodes
+        ok_pairs = filter (\(x, y) -> Node.idx x /= Node.idx y &&
+                                      Node.group x == Node.group y) all_pairs
+    in case count of
+         1 -> Ok (Left (map Node.idx all_nodes))
+         2 -> Ok (Right (map (\(p, s) -> (Node.idx p, Node.idx s)) ok_pairs))
+         _ -> Bad "Unsupported number of nodes, only one or two  supported"
 
 -- | Try to allocate an instance on the cluster.
 tryAlloc :: (Monad m) =>
             Node.List         -- ^ The node list
          -> Instance.List     -- ^ The instance list
          -> Instance.Instance -- ^ The instance to allocate
-         -> Int               -- ^ Required number of nodes
+         -> AllocNodes        -- ^ The allocation targets
          -> m AllocSolution   -- ^ Possible solution list
-tryAlloc nl _ inst 2 =
-    let all_nodes = getOnline nl
-        all_pairs = liftM2 (,) all_nodes all_nodes
-        ok_pairs = filter (\(x, y) -> Node.idx x /= Node.idx y) all_pairs
-        sols = foldl' (\cstate (p, s) ->
+tryAlloc nl _ inst (Right ok_pairs) =
+    let sols = foldl' (\cstate (p, s) ->
                            concatAllocs cstate $ allocateOnPair nl inst p s
-                      ) ([], 0, []) ok_pairs
-    in return sols
+                      ) emptySolution ok_pairs
 
-tryAlloc nl _ inst 1 =
-    let all_nodes = getOnline nl
-        sols = foldl' (\cstate ->
+    in if null ok_pairs -- means we have just one node
+       then fail "Not enough online nodes"
+       else return $ annotateSolution sols
+
+tryAlloc nl _ inst (Left all_nodes) =
+    let sols = foldl' (\cstate ->
                            concatAllocs cstate . allocateOnSingle nl inst
-                      ) ([], 0, []) all_nodes
-    in return sols
+                      ) emptySolution all_nodes
+    in if null all_nodes
+       then fail "No online nodes"
+       else return $ annotateSolution sols
 
-tryAlloc _ _ _ reqn = fail $ "Unsupported number of allocation \
-                             \destinations required (" ++ show reqn ++
-                                               "), only two supported"
+-- | Given a group/result, describe it as a nice (list of) messages
+solutionDescription :: Group.List -> (Gdx, Result AllocSolution) -> [String]
+solutionDescription gl (groupId, result) =
+  case result of
+    Ok solution -> map (printf "Group %s (%s): %s" gname pol) (asLog solution)
+    Bad message -> [printf "Group %s: error %s" gname message]
+  where grp = Container.find groupId gl
+        gname = Group.name grp
+        pol = apolToString (Group.allocPolicy grp)
 
--- | Try to allocate an instance on the cluster.
+-- | From a list of possibly bad and possibly empty solutions, filter
+-- only the groups with a valid result
+filterMGResults :: Group.List
+                -> [(Gdx, Result AllocSolution)]
+                -> [(Gdx, AllocSolution)]
+filterMGResults gl=
+  filter ((/= AllocUnallocable) . Group.allocPolicy .
+             flip Container.find gl . fst) .
+  filter (not . null . asSolutions . snd) .
+  map (\(y, Ok x) -> (y, x)) .
+  filter (isOk . snd)
+
+-- | Sort multigroup results based on policy and score
+sortMGResults :: Group.List
+             -> [(Gdx, AllocSolution)]
+             -> [(Gdx, AllocSolution)]
+sortMGResults gl sols =
+    let extractScore = \(_, _, _, x) -> x
+        solScore (gdx, sol) = (Group.allocPolicy (Container.find gdx gl),
+                               (extractScore . head . asSolutions) sol)
+    in sortBy (comparing solScore) sols
+
+-- | Try to allocate an instance on a multi-group cluster.
+tryMGAlloc :: Group.List           -- ^ The group list
+           -> Node.List            -- ^ The node list
+           -> Instance.List        -- ^ The instance list
+           -> Instance.Instance    -- ^ The instance to allocate
+           -> Int                  -- ^ Required number of nodes
+           -> Result AllocSolution -- ^ Possible solution list
+tryMGAlloc mggl mgnl mgil inst cnt =
+  let groups = splitCluster mgnl mgil
+      sols = map (\(gid, (nl, il)) ->
+                   (gid, genAllocNodes mggl nl cnt False >>=
+                       tryAlloc nl il inst))
+             groups::[(Gdx, Result AllocSolution)]
+      all_msgs = concatMap (solutionDescription mggl) sols
+      goodSols = filterMGResults mggl sols
+      sortedSols = sortMGResults mggl goodSols
+  in if null sortedSols
+     then Bad $ intercalate ", " all_msgs
+     else let (final_group, final_sol) = head sortedSols
+              final_name = Group.name $ Container.find final_group mggl
+              selmsg = "Selected group: " ++  final_name
+          in Ok $ final_sol { asLog = selmsg:all_msgs }
+
+-- | Try to relocate an instance on the cluster.
 tryReloc :: (Monad m) =>
             Node.List       -- ^ The node list
          -> Instance.List   -- ^ The instance list
@@ -596,73 +740,161 @@ tryReloc nl il xid 1 ex_idx =
                             let em = do
                                   (mnl, i, _, _) <-
                                       applyMove nl inst (ReplaceSecondary x)
-                                  return (mnl, i, [Container.find x mnl])
+                                  return (mnl, i, [Container.find x mnl],
+                                          compCV mnl)
                             in concatAllocs cstate em
-                       ) ([], 0, []) valid_idxes
+                       ) emptySolution valid_idxes
     in return sols1
 
 tryReloc _ _ _ reqn _  = fail $ "Unsupported number of relocation \
                                 \destinations required (" ++ show reqn ++
                                                   "), only one supported"
 
+tryMGReloc :: (Monad m) =>
+              Group.List      -- ^ The group list
+           -> Node.List       -- ^ The node list
+           -> Instance.List   -- ^ The instance list
+           -> Idx             -- ^ The index of the instance to move
+           -> Int             -- ^ The number of nodes required
+           -> [Ndx]           -- ^ Nodes which should not be used
+           -> m AllocSolution -- ^ Solution list
+tryMGReloc _ mgnl mgil xid ncount ex_ndx = do
+  let groups = splitCluster mgnl mgil
+      -- TODO: we only relocate inside the group for now
+      inst = Container.find xid mgil
+  (nl, il) <- case lookup (instancePriGroup mgnl inst) groups of
+                Nothing -> fail $ "Cannot find group for instance " ++
+                           Instance.name inst
+                Just v -> return v
+  tryReloc nl il xid ncount ex_ndx
+
+-- | Change an instance's secondary node
+evacInstance :: (Monad m) =>
+                [Ndx]                      -- ^ Excluded nodes
+             -> Instance.List              -- ^ The current instance list
+             -> (Node.List, AllocSolution) -- ^ The current state
+             -> Idx                        -- ^ The instance to evacuate
+             -> m (Node.List, AllocSolution)
+evacInstance ex_ndx il (nl, old_as) idx = do
+  -- FIXME: hardcoded one node here
+
+  -- Longer explanation: evacuation is currently hardcoded to DRBD
+  -- instances (which have one secondary); hence, even if the
+  -- IAllocator protocol can request N nodes for an instance, and all
+  -- the message parsing/loading pass this, this implementation only
+  -- supports one; this situation needs to be revisited if we ever
+  -- support more than one secondary, or if we change the storage
+  -- model
+  new_as <- tryReloc nl il idx 1 ex_ndx
+  case asSolutions new_as of
+    -- an individual relocation succeeded, we kind of compose the data
+    -- from the two solutions
+    csol@(nl', _, _, _):_ ->
+        return (nl', new_as { asSolutions = csol:asSolutions old_as })
+    -- this relocation failed, so we fail the entire evac
+    _ -> fail $ "Can't evacuate instance " ++
+         Instance.name (Container.find idx il) ++
+             ": " ++ describeSolution new_as
+
 -- | Try to evacuate a list of nodes.
 tryEvac :: (Monad m) =>
             Node.List       -- ^ The node list
          -> Instance.List   -- ^ The instance list
-         -> [Ndx]           -- ^ Nodes to be evacuated
+         -> [Idx]           -- ^ Instances to be evacuated
+         -> [Ndx]           -- ^ Restricted nodes (the ones being evacuated)
          -> m AllocSolution -- ^ Solution list
-tryEvac nl il ex_ndx =
+tryEvac nl il idxs ex_ndx = do
+  (_, sol) <- foldM (evacInstance ex_ndx il) (nl, emptySolution) idxs
+  return sol
+
+-- | Multi-group evacuation of a list of nodes.
+tryMGEvac :: (Monad m) =>
+             Group.List -- ^ The group list
+          -> Node.List       -- ^ The node list
+          -> Instance.List   -- ^ The instance list
+          -> [Ndx]           -- ^ Nodes to be evacuated
+          -> m AllocSolution -- ^ Solution list
+tryMGEvac _ nl il ex_ndx =
     let ex_nodes = map (`Container.find` nl) ex_ndx
         all_insts = nub . concatMap Node.sList $ ex_nodes
+        gni = splitCluster nl il
+        -- we run the instance index list through a couple of maps to
+        -- get finally to a structure of the type [(group index,
+        -- [instance indices])]
+        all_insts' = map (\idx ->
+                              (instancePriGroup nl (Container.find idx il),
+                               idx)) all_insts
+        all_insts'' = groupBy ((==) `on` fst) all_insts'
+        all_insts3 = map (\xs -> let (gdxs, idxs) = unzip xs
+                                 in (head gdxs, idxs)) all_insts''
     in do
-      (_, sol) <- foldM (\(nl', (_, _, rsols)) idx -> do
-                           -- FIXME: hardcoded one node here
-                           (fm, cs, aes) <- tryReloc nl' il idx 1 ex_ndx
-                           case aes of
-                             csol@(_, (nl'', _, _)):_ ->
-                                 return (nl'', (fm, cs, csol:rsols))
-                             _ -> fail $ "Can't evacuate instance " ++
-                                  show idx
-                        ) (nl, ([], 0, [])) all_insts
-      return sol
+      -- that done, we now add the per-group nl/il to the tuple
+      all_insts4 <-
+          mapM (\(gdx, idxs) -> do
+                  case lookup gdx gni of
+                    Nothing -> fail $ "Can't find group index " ++ show gdx
+                    Just (gnl, gil) -> return (gdx, gnl, gil, idxs))
+          all_insts3
+      results <- mapM (\(_, gnl, gil, idxs) -> tryEvac gnl gil idxs ex_ndx)
+                 all_insts4
+      let sol = foldl' (\orig_sol group_sol ->
+                        sumAllocs orig_sol group_sol) emptySolution results
+      return $ annotateSolution sol
 
 -- | Recursively place instances on the cluster until we're out of space
 iterateAlloc :: Node.List
              -> Instance.List
              -> Instance.Instance
-             -> Int
+             -> AllocNodes
              -> [Instance.Instance]
-             -> Result (FailStats, Node.List, [Instance.Instance])
-iterateAlloc nl il newinst nreq ixes =
+             -> [CStats]
+             -> Result AllocResult
+iterateAlloc nl il newinst allocnodes ixes cstats =
       let depth = length ixes
           newname = printf "new-%d" depth::String
           newidx = length (Container.elems il) + depth
           newi2 = Instance.setIdx (Instance.setName newinst newname) newidx
-      in case tryAlloc nl il newi2 nreq of
+      in case tryAlloc nl il newi2 allocnodes of
            Bad s -> Bad s
-           Ok (errs, _, sols3) ->
+           Ok (AllocSolution { asFailures = errs, asSolutions = sols3 }) ->
                case sols3 of
-                 [] -> Ok (collapseFailures errs, nl, ixes)
-                 (_, (xnl, xi, _)):[] ->
-                     iterateAlloc xnl il newinst nreq $! (xi:ixes)
+                 [] -> Ok (collapseFailures errs, nl, il, ixes, cstats)
+                 (xnl, xi, _, _):[] ->
+                     iterateAlloc xnl (Container.add newidx xi il)
+                                  newinst allocnodes (xi:ixes)
+                                  (totalResources xnl:cstats)
                  _ -> Bad "Internal error: multiple solutions for single\
                           \ allocation"
 
+-- | The core of the tiered allocation mode
 tieredAlloc :: Node.List
             -> Instance.List
             -> Instance.Instance
-            -> Int
+            -> AllocNodes
             -> [Instance.Instance]
-            -> Result (FailStats, Node.List, [Instance.Instance])
-tieredAlloc nl il newinst nreq ixes =
-    case iterateAlloc nl il newinst nreq ixes of
+            -> [CStats]
+            -> Result AllocResult
+tieredAlloc nl il newinst allocnodes ixes cstats =
+    case iterateAlloc nl il newinst allocnodes ixes cstats of
       Bad s -> Bad s
-      Ok (errs, nl', ixes') ->
+      Ok (errs, nl', il', ixes', cstats') ->
           case Instance.shrinkByType newinst . fst . last $
                sortBy (comparing snd) errs of
-            Bad _ -> Ok (errs, nl', ixes')
+            Bad _ -> Ok (errs, nl', il', ixes', cstats')
             Ok newinst' ->
-                tieredAlloc nl' il newinst' nreq ixes'
+                tieredAlloc nl' il' newinst' allocnodes ixes' cstats'
+
+-- | Compute the tiered spec string description from a list of
+-- allocated instances.
+tieredSpecMap :: [Instance.Instance]
+              -> [String]
+tieredSpecMap trl_ixes =
+    let fin_trl_ixes = reverse trl_ixes
+        ix_byspec = groupBy ((==) `on` Instance.specOf) fin_trl_ixes
+        spec_map = map (\ixs -> (Instance.specOf $ head ixs, length ixs))
+                   ix_byspec
+    in  map (\(spec, cnt) -> printf "%d,%d,%d=%d" (rspecMem spec)
+                             (rspecDsk spec) (rspecCpu spec) cnt) spec_map
 
 -- * Formatting functions
 
@@ -825,3 +1057,43 @@ iMoveToJob nl il idx move =
          ReplaceSecondary ns -> [ opR ns ]
          ReplaceAndFailover np -> [ opR np, opF ]
          FailoverAndReplace ns -> [ opF, opR ns ]
+
+-- * Node group functions
+
+-- | Computes the group of an instance
+instanceGroup :: Node.List -> Instance.Instance -> Result Gdx
+instanceGroup nl i =
+  let sidx = Instance.sNode i
+      pnode = Container.find (Instance.pNode i) nl
+      snode = if sidx == Node.noSecondary
+              then pnode
+              else Container.find sidx nl
+      pgroup = Node.group pnode
+      sgroup = Node.group snode
+  in if pgroup /= sgroup
+     then fail ("Instance placed accross two node groups, primary " ++
+                show pgroup ++ ", secondary " ++ show sgroup)
+     else return pgroup
+
+-- | Computes the group of an instance per the primary node
+instancePriGroup :: Node.List -> Instance.Instance -> Gdx
+instancePriGroup nl i =
+  let pnode = Container.find (Instance.pNode i) nl
+  in  Node.group pnode
+
+-- | Compute the list of badly allocated instances (split across node
+-- groups)
+findSplitInstances :: Node.List -> Instance.List -> [Instance.Instance]
+findSplitInstances nl il =
+  filter (not . isOk . instanceGroup nl) (Container.elems il)
+
+-- | Splits a cluster into the component node groups
+splitCluster :: Node.List -> Instance.List ->
+                [(Gdx, (Node.List, Instance.List))]
+splitCluster nl il =
+  let ngroups = Node.computeGroups (Container.elems nl)
+  in map (\(guuid, nodes) ->
+           let nidxs = map Node.idx nodes
+               nodes' = zip nidxs nodes
+               instances = Container.filter ((`elem` nidxs) . Instance.pNode) il
+           in (guuid, (Container.fromList nodes', instances))) ngroups
