@@ -4,7 +4,7 @@
 
 {-
 
-Copyright (C) 2009 Google Inc.
+Copyright (C) 2009, 2010 Google Inc.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -26,9 +26,14 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 module Ganeti.HTools.Types
     ( Idx
     , Ndx
+    , Gdx
     , NameAssoc
     , Score
     , Weight
+    , GroupID
+    , AllocPolicy(..)
+    , apolFromString
+    , apolToString
     , RSpec(..)
     , DynUtil(..)
     , zeroUtil
@@ -46,6 +51,8 @@ module Ganeti.HTools.Types
     , MoveJob
     , JobSet
     , Result(..)
+    , isOk
+    , isBad
     , Element(..)
     , FailMode(..)
     , FailStats
@@ -54,14 +61,20 @@ module Ganeti.HTools.Types
     , queryTimeout
     ) where
 
+import qualified Data.Map as M
+import qualified Text.JSON as JSON
+
 -- | The instance index type.
 type Idx = Int
 
 -- | The node index type.
 type Ndx = Int
 
+-- | The group index type.
+type Gdx = Int
+
 -- | The type used to hold name-to-idx mappings.
-type NameAssoc = [(String, Int)]
+type NameAssoc = M.Map String Int
 
 -- | A separate name for the cluster score type.
 type Score = Double
@@ -69,12 +82,52 @@ type Score = Double
 -- | A separate name for a weight metric.
 type Weight = Double
 
+-- | The Group UUID type.
+type GroupID = String
+
+-- | The Group allocation policy type.
+--
+-- Note that the order of constructors is important as the automatic
+-- Ord instance will order them in the order they are defined, so when
+-- changing this data type be careful about the interaction with the
+-- desired sorting order.
+data AllocPolicy
+    = AllocPreferred   -- ^ This is the normal status, the group
+                       -- should be used normally during allocations
+    | AllocLastResort  -- ^ This group should be used only as
+                       -- last-resort, after the preferred groups
+    | AllocUnallocable -- ^ This group must not be used for new
+                       -- allocations
+      deriving (Show, Read, Eq, Ord)
+
+-- | Convert a string to an alloc policy
+apolFromString :: (Monad m) => String -> m AllocPolicy
+apolFromString s =
+    case s of
+      "preferred"   -> return AllocPreferred
+      "last_resort" -> return AllocLastResort
+      "unallocable" -> return AllocUnallocable
+      o -> fail $ "Invalid alloc policy mode: " ++ o
+
+-- | Convert an alloc policy to the Ganeti string equivalent
+apolToString :: AllocPolicy -> String
+apolToString AllocPreferred   = "preferred"
+apolToString AllocLastResort  = "last_resort"
+apolToString AllocUnallocable = "unallocable"
+
+instance JSON.JSON AllocPolicy where
+    showJSON = JSON.showJSON . apolToString
+    readJSON s = case JSON.readJSON s of
+                   JSON.Ok s' -> apolFromString s'
+                   JSON.Error e -> JSON.Error $
+                                   "Can't parse alloc_policy: " ++ e
+
 -- | The resource spec type.
 data RSpec = RSpec
     { rspecCpu  :: Int  -- ^ Requested VCPUs
     , rspecMem  :: Int  -- ^ Requested memory
     , rspecDsk  :: Int  -- ^ Requested disk
-    } deriving (Show, Eq)
+    } deriving (Show, Read, Eq)
 
 -- | The dynamic resource specs of a machine (i.e. load or load
 -- capacity, as opposed to size).
@@ -83,7 +136,7 @@ data DynUtil = DynUtil
     , memWeight :: Weight -- ^ Standardised memory load
     , dskWeight :: Weight -- ^ Standardised disk I\/O usage
     , netWeight :: Weight -- ^ Standardised network usage
-    } deriving (Show)
+    } deriving (Show, Read, Eq)
 
 -- | Initial empty utilisation
 zeroUtil :: DynUtil
@@ -113,7 +166,7 @@ data IMove = Failover                -- ^ Failover the instance (f)
            | ReplaceSecondary Ndx    -- ^ Replace secondary (r:ns)
            | ReplaceAndFailover Ndx  -- ^ Replace secondary, failover (r:np, f)
            | FailoverAndReplace Ndx  -- ^ Failover, replace secondary (f, r:ns)
-             deriving (Show)
+             deriving (Show, Read)
 
 -- | Formatted solution output for one move (involved nodes and
 -- commands
@@ -164,7 +217,7 @@ mini-library here
 data Result a
     = Bad String
     | Ok a
-    deriving (Show)
+    deriving (Show, Read)
 
 instance Monad Result where
     (>>=) (Bad x) _ = Bad x
@@ -172,13 +225,22 @@ instance Monad Result where
     return = Ok
     fail = Bad
 
+-- | Simple checker for whether Result is OK
+isOk :: Result a -> Bool
+isOk (Ok _) = True
+isOk _ = False
+
+-- | Simple checker for whether Result is a failure
+isBad :: Result a  -> Bool
+isBad = not . isOk
+
 -- | Reason for an operation's falure
 data FailMode = FailMem  -- ^ Failed due to not enough RAM
               | FailDisk -- ^ Failed due to not enough disk
               | FailCPU  -- ^ Failed due to not enough CPU capacity
               | FailN1   -- ^ Failed due to not passing N1 checks
               | FailTags -- ^ Failed due to tag exclusion
-                deriving (Eq, Enum, Bounded, Show)
+                deriving (Eq, Enum, Bounded, Show, Read)
 
 -- | List with failure statistics
 type FailStats = [(FailMode, Int)]
@@ -186,7 +248,7 @@ type FailStats = [(FailMode, Int)]
 -- | Either-like data-type customized for our failure modes
 data OpResult a = OpFail FailMode -- ^ Failed operation
                 | OpGood a        -- ^ Success operation
-                  deriving (Show)
+                  deriving (Show, Read)
 
 instance Monad OpResult where
     (OpGood x) >>= fn = fn x
